from colored.colored import bg, fg

from a9s.aws_resources.base_service import BaseService
from a9s.aws_resources.utils import pop_if_exists
from a9s.components.custom_string import String
from a9s.components.keys import BACK_KEYS, is_match
from a9s.components.logger import logger
from a9s.components.table import ColSettings, updates_table_data_method


class Route53Table(BaseService):
    SERVICE_NAME = 'Route 53'
    BOTO_SERVICE = 'route53'

    def __init__(self) -> None:
        self.hosted_zone = None
        self._selection_stack = []
        self._filter_stack = []

        super().__init__([], [])

    def initialize(self):
        self.queue_action(self.list_hosted_zones, self.on_updated_data)

    def get_hud_text(self, space_left):
        if not self.hosted_zone:
            return super().get_hud_text(space_left)

        return String(self.hosted_zone['Name'], bg=bg('medium_purple_2a'), fg=fg('black')).reset_style_on_end()

    def handle_key(self, key):
        should_stop_propagate = super().handle_key(key)
        if is_match(key, BACK_KEYS) and not should_stop_propagate:
            if self.filter:
                return should_stop_propagate

            if self.hosted_zone is not None:
                filter_str = pop_if_exists(self._filter_stack, default='')
                selected_row = pop_if_exists(self._selection_stack, default=0)
                logger.debug(f'Popped {filter_str} and {selected_row} from stack')
                self.queue_action(self.list_hosted_zones, self.on_updated_data, filter_str=filter_str, selected_row=selected_row)
                self.hosted_zone = None
                should_stop_propagate = True

        return should_stop_propagate

    def on_select(self, data):
        if not self.hosted_zone:
            self._filter_stack.append(self.filter)
            self._selection_stack.append(self.selected_row)
            logger.debug(f'Pushing {self.filter} and {self.selected_row} to stack')
            self.hosted_zone = data
            self.queue_action(self.list_records, self.on_updated_data)

    @updates_table_data_method
    def list_hosted_zones(self):
        response = self.client.list_hosted_zones()
        headers = [ColSettings("Name", yank_key='n', stretched=True, min_size=20), ColSettings("ID", yank_key='i'), ColSettings("Records")]

        data = []
        for item in response['HostedZones']:
            data.append([item['Name'], item['Id'], str(item['ResourceRecordSetCount'])])

        return headers, data

    @updates_table_data_method
    def list_records(self):
        headers = [ColSettings('Name', yank_key='n', min_size=20, stretched=True), ColSettings('Type'), ColSettings('TTL', yank_key='t'), ColSettings('Record', yank_key='r')]
        data = []

        should_continue = True
        next_record_name = None
        while should_continue:
            if next_record_name is None:
                response = self.client.list_resource_record_sets(HostedZoneId=self.hosted_zone['ID'], MaxItems='300')
            else:
                response = self.client.list_resource_record_sets(HostedZoneId=self.hosted_zone['ID'], StartRecordName=next_record_name, MaxItems='300')

            if not response['IsTruncated']:
                should_continue = False

            else:
                next_record_name = response['NextRecordName']

            for item in response['ResourceRecordSets']:
                record = item.get('AliasTarget', {}).get('DNSName')
                if not record and item.get('ResourceRecords'):
                    record = ', '.join(r['Value'] for r in item['ResourceRecords'])
                data.append([item['Name'], item['Type'], str(item.get('TTL', '')), record])

        return headers, data



from a9s.components.renderer import Renderer

LOGO = [
    "      ________       ",
    "_____/   __   \______",
    "\__  \____    /  ___/",
    " / __ \_/    /\___ \ ",
    "(____  /____//____  >",
    "     \/           \/ "
]



class Logo(Renderer):
    def __init__(self, version):
        super(Logo, self).__init__()
        self.version = version

    def draw(self):
        for line in LOGO:
            self.echo(line)

        version_text = self.version
        version_x_pos = round(self.width / 2 - len(version_text) / 2 + self.x)
        self.echo(version_text, x=version_x_pos)



from typing import Callable, Dict

from colored import attr, fg

from a9s.components.custom_string import String
from a9s.components.keys import RIGHT_KEYS, is_match
from a9s.components.mode import KeyMode
from a9s.components.renderer import Renderer


class AutoComplete(Renderer):
    def __init__(self, commands: Dict[str, Callable]):
        super(AutoComplete, self).__init__()
        self._index = 0
        self._buffer = [" "] * self.width
        self.text = ""

        self.commands = commands
        self.guess_mode = KeyMode.Navigation

    def delete_char(self):
        if len(self.text) > 1:
            self.text = self.text[:-1]

    def get_actual_text(self):
        return self.text[1:]

    def execute_command_if_matched(self):
        for command, func in self.commands.items():
            if command.startswith(self.get_actual_text()):
                return func()

    def handle_key(self, key):
        guessed_command = self.guess_text()
        if is_match(key, RIGHT_KEYS, code_only=True):
            if guessed_command:
                self.text = ":" + guessed_command

    def guess_text(self):
        text = self.get_actual_text()
        if len(text) == 0:
            return ""

        if self.guess_mode == KeyMode.Command:
            for command in self.commands:
                if command.startswith(text.lower()):
                    return command

        return ""

    def draw(self):
        guessed_command = self.guess_text()
        text = self.get_actual_text()
        unmatched_part = guessed_command[len(text):]

        to_print = (String(self.text, fg=attr('reset')) + String(unmatched_part, fg=fg('grey_50'))).pad(self.width)[-self.width:]

        self.echo(to_print)



from colored import fg, bg

from a9s.components.custom_string import Style
from a9s.components.keys import BACK_KEYS, is_match
from a9s.components.renderer import ScrollableRenderer

MAX_RECORDS = 100


class Logger(ScrollableRenderer):
    def __init__(self):
        super(Logger, self).__init__([])
        self._halt_debug = False
        self.shown = False

        self.halt_buffer = []

    def halt_debug(self):
        self._halt_debug = True

    def continue_debug(self):
        for msg in self.halt_buffer:
            self._add_message(msg)

        self.halt_buffer = []
        self._halt_debug = False

    def _add_message(self, msg: str):
        self.data = (self._data + [msg])[-MAX_RECORDS:]

    def on_data_set(self, new_value):
        self._displayed_data_start = max(len(self.displayed_data) - self.height + 1, 0)
        self.selected_row = len(self.displayed_data) - 1

    def debug(self, msg: str):
        if self._halt_debug:
            self.halt_buffer.append(msg)
            return

        self._add_message(msg)

    def handle_key(self, key) -> bool:
        should_stop_propagate = super(Logger, self).handle_key(key)
        if not should_stop_propagate:
            if is_match(key, BACK_KEYS):
                self.continue_debug()

        return should_stop_propagate

    def draw(self):
        for i, log in enumerate(self.displayed_data[self.displayed_data_start:self.displayed_data_end]):
            actual_i = self.displayed_data_start + i
            row_style = Style()
            if self.selected_row == actual_i:
                row_style = Style(fg=fg("black"), bg=bg("light_gray"))

            log = log + " " * (self.width - len(log))
            self.echo(self._trim_row(log).with_style(row_style))


logger = Logger()



import os

import boto3
from cached_property import cached_property

from a9s.aws_resources.hud import HUDComponent
from a9s.components.table import Table


IS_LOCAL = os.environ.get('LOCAL', 'false').lower() == 'true'


class BaseService(Table, HUDComponent):
    BOTO_SERVICE = None

    @cached_property
    def client(self):
        return boto3.client(service_name=self.BOTO_SERVICE, endpoint_url='http://localhost:54321' if IS_LOCAL else None)



import itertools
import pyperclip
from copy import copy
from dataclasses import dataclass, field
from typing import Union, Callable, List

from colored import fg, bg, attr

from a9s.components.renderer import ScrollableRenderer
from a9s.components.custom_string import String, Style


@dataclass
class ColSettings:
    name: str
    maxWidth: int = None
    stretched: bool = False
    padding: int = 3
    min_size: Union[int, Callable[[int, "ColSettings"], int]] = None
    max_size: Union[int, Callable[[int, "ColSettings"], int]] = None
    no_shrink: bool = False
    yank_key: str = None

    data_max_len: int = field(init=False, default=0)
    actual_size: int = field(init=False, default=0)
    _min_size: int = 0
    _max_size: int = 0

    def __post_init__(self):
        self.user_defined_min_size = self.min_size is not None
        self.user_defined_max_size = self.max_size is not None

    def enrich_props_using_data(self, col):
        max_data_size = max(len(value) if value else 0 for value in col)
        self.data_max_len = max_data_size

        self._min_size = max(min(len(self.name), max_data_size), 3)
        if self.min_size:
            if callable(self.min_size):
                val = self.min_size(max_data_size, self)

            else:
                val = self.min_size

            self._min_size = max(val, self._min_size)

        self._max_size = max(max_data_size, len(self.name), self._min_size)
        if self.max_size:
            if callable(self.max_size):
                val = self.max_size(max_data_size, self)

            else:
                val = self.max_size

            self._max_size = max(min(val, self._max_size), self._min_size)

        self.actual_size = self._max_size

    def get_full_size(self):
        return self.actual_size + self.padding * 2

    def format_substr(self, data, search, match_style, row_style):
        if len(search) == 0:
            return data

        fragments = data.split(search, 1)
        to_append = String(search, fg=match_style).force_self_style()
        return to_append.join(fragments)

    def format_data(self, data, search="", row_style=attr("reset"), match_style=fg("blue"), yank_mode=False):
        if isinstance(data, str):
            data = String(data)
        
        data = copy(data)

        used_max_size = self.actual_size
        match_index = -1
        if len(search) > 0:
            match_index = data.find(search)

        to_ret = data
        if match_index != -1:
            to_ret = self.format_substr(data, search, match_style, row_style)
        if len(data) > self.actual_size:
            if self.actual_size > self._min_size:
                to_ret = to_ret[:self.actual_size - 3]
                if match_index > self.actual_size - 3 or (match_index != -1 and len(search) > self.actual_size - 3):
                    to_ret += String("...", fg=match_style)
                else:
                    to_ret += "..."

            else:
                if len(data) > self._min_size:
                    to_ret = to_ret[:self._min_size - 3]
                    if match_index > self._min_size - 3 or (match_index != -1 and len(search) > self._min_size - 3):
                        to_ret += String("...", fg=match_style)
                    else:
                        to_ret += "..."
                else:
                    to_ret = to_ret[:self._min_size]

                used_max_size = self._min_size


        if yank_mode and self.yank_key:
            to_ret += String("({})".format(self.yank_key))
    
        data_len = len(to_ret)
        if used_max_size - data_len > 0:
            to_ret += " " * (used_max_size - data_len)

        return to_ret


class Table(ScrollableRenderer):
    def __init__(self, headers: List[ColSettings], data):
        super(Table, self).__init__(data)
        self._transposed_data = []

        self.headers = headers
        self.data = data
        self.data_updating = True
    
    def _on_select(self, data):
        if not self.data_updating:
            super(Table, self)._on_select(data)

    def on_updated_data(self, data, filter_str='', selected_row=0):
        self.headers, self.data = data
        self.filter = filter_str
        self.selected_row = selected_row
        self.data_updating = False

    def on_data_set(self, new_value):
        self._transposed_data = list(map(list, itertools.zip_longest(*new_value, fillvalue=None)))
        self.selected_row = 0
        self.offset = 0
        self._displayed_data_start = 0
        for header, col in zip(self.headers, self._transposed_data):
            header.enrich_props_using_data(col)
    
    def handle_key(self, key):
        should_stop_propagate = super(Table, self).handle_key(key)
        if not should_stop_propagate and self.yank_mode:
            for header in self.headers:
                if key == header.yank_key:
                    pyperclip.copy(self.currently_selected_data[header.name])
                    self.yank_mode = False
                    should_stop_propagate = True

        return should_stop_propagate or self.data_updating

    def _enrich_reponsive_headers(self):
        stretched_headers = []
        stretched_size = self.width
        stretched_padding = 0
        for header in self.headers:
            if header.stretched:
                stretched_padding += header.padding * 2
                stretched_headers.append(header)
                continue
            stretched_size -= header.get_full_size()

        stretched_size -= stretched_padding
        for header in stretched_headers:
            if header.user_defined_max_size and stretched_size > header._max_size:
                continue

            if header.user_defined_min_size and stretched_size < header._min_size:
                continue

            header.actual_size = stretched_size
        return stretched_size

    @property
    def filtered_data(self):
        if len(self.filter) == 0:
            return self.data

        to_ret = []
        for data_row in self.data:
            found = False
            for col in data_row:
                if self.filter and self.filter in col:
                    found = True
                    break

            if found:
                to_ret.append(data_row)

        return to_ret

    def get_row_representation(self, data):
        to_ret = {}
        for col, d in zip(self.headers, data):
            if isinstance(d, String):
                d = repr(d)
            to_ret[col.name] = d

        return to_ret

    def draw(self):
        header_style = Style(fg=fg("light_gray") + attr('bold'), bg=bg("dodger_blue_2"))
        row_to_print = String("").with_style(header_style)
        self._enrich_reponsive_headers()

        for header in self.headers:
            padding = " " * header.padding
            row_to_print += String(padding).with_style(header_style)
            row_to_print += header.format_data(header.name.upper(), yank_mode=self.yank_mode)
            row_to_print += String(padding).with_style(header_style)

        self.max_offset = len(row_to_print) - self.width
        row_to_print = self._trim_row(row_to_print, with_elipsis=True)
        self.echo(row_to_print)

        for i, data_row in enumerate(self.displayed_data[self.displayed_data_start:self.displayed_data_end]):
            actual_i = self.displayed_data_start + i
            row_style = Style()
            if self.selected_row == actual_i:
                row_style = Style(fg=fg("black"), bg=bg("light_gray"))

            row_to_print = String("").with_style(row_style)
            for header, data in zip(self.headers, data_row):
                padding = " " * header.padding
                row_to_print += String(padding).with_style(row_style)
                row_to_print += header.format_data(data, self._filter)
                row_to_print += String(padding).with_style(row_style)

            row_to_print = self._trim_row(row_to_print)
            self.echo(row_to_print)


def updates_table_data_method(f):
    def inner(self, *args, **kwargs):
        self.data_updating = True
        return f(self, *args, **kwargs)

    return inner



import asyncio

from cached_property import cached_property
from typing import Union

from colored import attr

from a9s.components.custom_string import String
from a9s.components.keys import RIGHT_KEYS, is_match, LEFT_KEYS, UP_KEYS, DOWN_KEYS, ENTER_KEYS, BACK_KEYS, END_OF_LINE_KEYS, START_OF_LINE_KEYS, \
    START_OF_DOC_KEYS, END_OF_DOC_KEYS, COPY_KEYS
from a9s.components.utils import FPS


class Renderer:
    hidden_cursor: bool = True

    def __init__(self):
        self.x = 0
        self.y = 0
        self.to_x = self.x + 1
        self.to_y = self.y

        # renderer row
        self._curr_row = 0
        self._curr_col = 0

        self._echo_func = None
        self._async_coroutines = {}
        self._pending_coroutines = []
        self.force_empty_fill = False

        self._shown = True

    def initialize(self):
        pass

    @property
    def shown(self):
        return self._shown

    @shown.setter
    def shown(self, value):
        if value:
            self.show()

        else:
            self.hide()

    def hide(self):
        self.force_empty_fill = True
        self._shown = False

    def show(self):
        self._shown = True

    @property
    def width(self):
        return self.to_x - self.x

    @property
    def height(self):
        return self.to_y - self.y

    def fill_empty(self):
        try:
            if not self.shown and not self.force_empty_fill:
                return

            filler = String(" " * (self.width - self._curr_col), fg=attr('reset'))
            while self._curr_row < self.height:
                self.echo(filler)

        finally:
            self.set_relative_pos(0, 0)

    def set_relative_pos(self, x=None, y=None):
        if x is not None:
            self._curr_col = x

        if y is not None:
            self._curr_row = y

    def echo(self, string: Union[String, str], *, x=None, y=None, no_new_line=False):
        if not self._echo_func:
            raise RuntimeError('Echo func was not defined!')

        y_defined = y is not None
        y = self.y + self._curr_row if y is None else y

        x = self.x + self._curr_col if x is None else x
        self._echo_func(x, y, string)
        if not y_defined and not no_new_line:
            self._curr_row += 1

    async def update_data(self):
        captured_routines = self._pending_coroutines
        self._pending_coroutines = []
        for coroutine in captured_routines:
            task = asyncio.create_task(asyncio.to_thread(coroutine['routine']))
            self._async_coroutines[task] = coroutine

        coroutines = set(self._async_coroutines.keys())
        if len(coroutines) == 0:
            return

        done, pending = await asyncio.wait(coroutines, timeout=FPS)
        for routine in done:
            routine_describe = self._async_coroutines[routine]
            handler = routine_describe['handler']
            handler(routine.result(), *routine_describe['args'], **routine_describe['kwargs'])
            del self._async_coroutines[routine]

    def queue_action(self, routine, callback, *args, **kwargs):
        self._pending_coroutines.append(dict(handler=callback, routine=routine, args=args, kwargs=kwargs))

    def draw(self):
        pass

    def render(self, echo_func):
        self._echo_func = echo_func
        if self.shown:
            self.draw()

        self.fill_empty()

    def set_pos(self, *, x, y, to_x, to_y=None):
        self.x = x
        self.y = y
        self.to_x = to_x
        self.to_y = to_y if to_y else y + 1

    def handle_key(self, key):
        pass

    def onresize(self):
        pass


class ScrollableRenderer(Renderer):
    def __init__(self, data):
        super(ScrollableRenderer, self).__init__()
        self._data = data
        self._displayed_data = self._data

        self.offset = 0
        self.max_offset = 0
        self.selected_row = 0
        self._displayed_data_start = 0

        self._filter = ""
        self.filter = ""
        self.yank_mode = False

        self.data_updating = False

    def onresize(self):
        self.offset = 0

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new_value):
        self._data = new_value
        self.filter = ""
        if 'displayed_data' in self.__dict__:
            del self.__dict__['displayed_data']
        self.on_data_set(new_value)

    def on_data_set(self, new_value):
        pass

    @cached_property
    def displayed_data(self):
        return self.filtered_data

    @property
    def filter(self):
        return self._filter

    @filter.setter
    def filter(self, value):
        if self._filter != value:
            self.selected_row = 0
            self._displayed_data_start = 0

        self._filter = value
        if 'displayed_data' in self.__dict__:
            del self.__dict__['displayed_data']

    @property
    def filtered_data(self):
        if len(self.filter) == 0:
            return self.data

        to_ret = []
        for data_row in self.data:
            if self.filter and self.filter in data_row:
                to_ret.append(data_row)

        return to_ret

    def hscroll_end(self):
        self.offset = max(self.max_offset, 0)

    def hscroll_start(self):
        self.offset = 0

    def vscroll_start(self):
        self._displayed_data_start = 0
        self.selected_row = 0

    def vscroll_end(self):
        self._displayed_data_start = max(len(self.displayed_data) - self.height + 1, 0)
        self.selected_row = len(self.displayed_data) - 1

    @property
    def displayed_data_start(self):
        return self._displayed_data_start

    @property
    def displayed_data_end(self):
        return self.height + self.displayed_data_start - 1

    def move_right(self):
        if self.offset < self.max_offset:
            self.offset += 1

    def move_left(self):
        if self.offset > 0:
            self.offset -= 1

    def prev_selection(self):
        if self.selected_row > 0:
            self.selected_row -= 1

            if self.selected_row < self.displayed_data_start:
                self._displayed_data_start -= 1

    def next_selection(self):
        if self.selected_row < len(self.displayed_data) - 1:
            self.selected_row += 1

            if self.selected_row > self.displayed_data_end - 1:
                self._displayed_data_start += 1

    def get_row_representation(self, data):
        if isinstance(data, String):
            data = repr(data)

        return data
    
    def _on_select(self, data):
        self.on_select(data)

    def on_select(self, data):
        pass

    @property
    def currently_selected_data(self):
        data = self.displayed_data[self.selected_row]
        return self.get_row_representation(data)

    def handle_key(self, key) -> bool:
        if is_match(key, RIGHT_KEYS):
            self.move_right()

        elif is_match(key, LEFT_KEYS):
            self.move_left()

        if is_match(key, UP_KEYS):
            self.prev_selection()

        elif is_match(key, DOWN_KEYS):
            self.next_selection()
        
        elif is_match(key, ENTER_KEYS):
            self._on_select(self.currently_selected_data)
        
        elif is_match(key, BACK_KEYS):
            if self.yank_mode:
                self.yank_mode = False
                return True  # swallows the Exit key

            elif self.filter:
                self.filter = ''
                return True

        if is_match(key, END_OF_LINE_KEYS):
            self.hscroll_end()

        elif is_match(key, START_OF_LINE_KEYS):
            self.hscroll_start()

        elif is_match(key, START_OF_DOC_KEYS):
            self.vscroll_start()

        elif is_match(key, END_OF_DOC_KEYS):
            self.vscroll_end()
        
        elif is_match(key, COPY_KEYS):
            self.yank_mode = True

        return False

    def _trim_row(self, row, with_elipsis=False):
        actual_start = self.offset
        prefix = ""
        if self.offset > 0:
            prefix = "<- " if with_elipsis else "   "
            actual_start = len(prefix) + self.offset

        actual_end = self.offset + self.width
        suffix = ""
        if len(row) - self.offset > self.width:
            suffix = " ->" if with_elipsis else "   "
            actual_end = actual_end - len(suffix)

        return String(prefix, fg=attr("reset")) + row[actual_start:actual_end] + String(suffix, fg=attr("reset"))



from colored import fg, attr

from a9s.components.custom_string import String
from a9s.components.keys import SEARCH_KEYS, BACK_KEYS, COPY_KEYS, COMMAND_KEYS, END_OF_DOC_KEYS, START_OF_DOC_KEYS, START_OF_LINE_KEYS, END_OF_LINE_KEYS, \
    RIGHT_KEYS, LEFT_KEYS, DOWN_KEYS, UP_KEYS
from a9s.components.renderer import Renderer


NAV_COLUMN = [LEFT_KEYS, DOWN_KEYS, UP_KEYS, RIGHT_KEYS, START_OF_LINE_KEYS, END_OF_LINE_KEYS, START_OF_DOC_KEYS, END_OF_DOC_KEYS]


class Help(Renderer):
    COL_WIDTH = 25
    def __init__(self):
        super(Help, self).__init__()
        self.yank_mode = False

    def draw(self):
        command_style = fg('orchid_1') + attr('bold')
        key_style = fg('dodger_blue_2') + attr('bold')
        special_style = fg('medium_orchid') + attr('bold')
        self.echo(String("Keys:", fg=fg('orange_1') + attr('bold')))

        self.echo(String(" ") + SEARCH_KEYS.format(command_style))
        self.echo(String(" ") + COMMAND_KEYS.format(command_style))
        self.echo(String(" ") + BACK_KEYS.format(command_style))
        self.echo("-" * self.COL_WIDTH)
        yank_text = String(" ") + COPY_KEYS.format(special_style)
        if self.yank_mode:
            yank_text += String(" (ON)")

        else:
            yank_text.pad(self.COL_WIDTH)

        self.echo(yank_text)

        self.set_relative_pos(x=self.COL_WIDTH, y=1)
        for key_set in NAV_COLUMN:
            self.echo(String(" ") + key_set.format(key_style))



import os
from subprocess import call
import pathlib

import tempfile
from colored.colored import bg, fg

from a9s.aws_resources.base_service import BaseService
from a9s.aws_resources.utils import pop_if_exists
from a9s.components.custom_string import String
from a9s.components.keys import BACK_KEYS, is_match
from a9s.components.logger import logger
from a9s.components.table import ColSettings, updates_table_data_method
from a9s.components.app import EDITOR

IS_LOCAL = os.environ.get('LOCAL', 'false').lower() == 'true'


class S3Table(BaseService):
    SERVICE_NAME = 'S3'
    BOTO_SERVICE = 's3'

    def __init__(self) -> None:
        self.bucket = None
        self.paths = []
        self._selection_stack = []
        self._filter_stack = []

        super().__init__([], [])

    def initialize(self):
        self.queue_action(self.list_buckets, self.on_updated_data)

    def get_hud_text(self, space_left):
        if not self.bucket:
            return super().get_hud_text(space_left)

        return String([String(self.bucket, bg=bg('orange_1'), fg=fg('black')).reset_style_on_end(),
                       String("/"), String(self.prefix[-(space_left - len(self.bucket) - 1):])])
    
    @property
    def prefix(self):
        return "".join(self.paths)

    def handle_key(self, key):
        should_stop_propagate = super().handle_key(key)
        if is_match(key, BACK_KEYS) and not should_stop_propagate:
            filter_str = pop_if_exists(self._filter_stack, default='')
            selected_row = pop_if_exists(self._selection_stack, default=0)
            logger.debug(f'Popped {filter_str} and {selected_row} from stack')

            if len(self.paths) > 0:
                self.paths.pop()
                self.queue_action(self.list_bucket, self.on_updated_data, filter_str=filter_str, selected_row=selected_row)
                should_stop_propagate = True

            elif self.bucket is not None:
                self.queue_action(self.list_buckets, self.on_updated_data, filter_str=filter_str, selected_row=selected_row)
                self.bucket = None
                should_stop_propagate = True
        
        return should_stop_propagate
 
    def on_select(self, data):
        if not self.bucket:
            logger.debug(f'Pushing {self.filter} and {self.selected_row} to stack')
            self._filter_stack.append(self.filter)
            self._selection_stack.append(self.selected_row)
            self.bucket = data['Bucket name']
            self.queue_action(self.list_bucket, self.on_updated_data)

        else:
            if data['Type'] == "folder":
                logger.debug(f'Pushing {self.filter} and {self.selected_row} to stack')
                self._filter_stack.append(self.filter)
                self._selection_stack.append(self.selected_row)
                self.paths.append(data['Key'])
                self.queue_action(self.list_bucket, self.on_updated_data)

            else:  # we are a file
                full_key = self.prefix + data['Key']
                resp = self.client.get_object(Bucket=self.bucket, Key=full_key)
                with tempfile.NamedTemporaryFile(suffix="." + full_key.replace("/", "___")) as tf:
                    for chunk in resp['Body'].iter_chunks():
                        tf.write(chunk)

                    tf.flush()
                    call([EDITOR, tf.name])

    @updates_table_data_method
    def list_buckets(self):
        response = self.client.list_buckets()
        headers = [ColSettings("Bucket name", stretched=True, yank_key='n'), ColSettings("Creation date", yank_key='d')]
        data = []
        for bucket in response['Buckets']:
            data.append([bucket['Name'], str(bucket['CreationDate'])])

        return headers, data

    @updates_table_data_method
    def list_bucket(self):
        objects = self.client.list_objects_v2(Bucket=self.bucket, Delimiter="/", Prefix=self.prefix)
        headers = [ColSettings("Key", stretched=True, min_size=15, yank_key='k'), ColSettings("Type"), ColSettings("Last modify"), ColSettings("ETag", yank_key='t'), ColSettings("Size"), ColSettings("Storage Class"), ColSettings("Owner")]
        data = []
        for object in objects.get('CommonPrefixes', []):
            folder_data = [String(str(object['Prefix'].replace(self.prefix, "", 1)), fg=fg('magenta_3c')), "folder"]
            folder_data += (len(headers) - len(folder_data)) * [""]
            data.append(folder_data)

        for object in objects.get('Contents', []):
            name = object['Key'].replace(self.prefix, "", 1)
            file_type = "file ({})".format(pathlib.Path(name).suffix) if pathlib.Path(name).suffix else "file"
            data.append([name, file_type, str(object['LastModified']), object['ETag'], str(object['Size']), object['StorageClass'], object.get('Owner', {}).get('DisplayName', "")])

        return headers, data



import os
import sys
import asyncio
import tempfile
from typing import List
from subprocess import call

from colored import attr
from blessed import Terminal

from a9s.components.custom_string import String
from a9s.components.renderer import Renderer
from a9s.components.utils import FPS

EDITOR = os.environ.get('EDITOR', 'vim')


def flush():
    sys.stdout.flush()


class App:
    def __init__(self, render_list):
        self.term = Terminal()
        self.render_list: list[Renderer] = render_list
        self.should_run = True

        self.buffer = []
        self.on_resize()

    def on_resize(self, *args):
        self.buffer = [[' ' for x in range(self.term.width)] for y in range(self.term.height)]
        for rendered in self.render_list:
            rendered.onresize()
        sys.stdout.write(self.term.clear)

    def dump_to_screen(self):
        for line_num, line in enumerate(self.buffer):
            sys.stdout.write(self.term.move_xy(0, line_num) + "".join(line) + attr("reset"))

        sys.stdout.write(attr("reset"))

    def echo(self, x, y, to_print):
        if isinstance(to_print, str):
            to_print = String(to_print)

        if y >= len(self.buffer) or x + len(to_print) > len(self.buffer[0]):
            return

        line = self.buffer[y]
        to_print.reset_style_on_end()
        for i, char in enumerate(to_print):
            actual_x = x + i
            line[actual_x] = str(char)

    def add_to_render(self, renderer: Renderer):
        self.render_list.append(renderer)

    def add_multiple(self, renders: List[Renderer]):
        self.render_list.extend(renders)

    @classmethod
    def get_editor_input(cls, initial_text: str = ""):
        with tempfile.NamedTemporaryFile(suffix=".tmp") as tf:
            tf.write(initial_text.encode())
            tf.flush()
            call([EDITOR, tf.name])
            tf.seek(0)
            return tf.read().decode()

    async def draw_loop(self):
        with self.term.fullscreen():
            while self.should_run:
                with self.term.hidden_cursor():
                    await asyncio.gather(asyncio.sleep(FPS), *[renderer.update_data() for renderer in self.render_list])
                    for renderer in self.render_list:
                        renderer.render(self.echo)

                    self.dump_to_screen()
                    flush()

    async def input_loop(self):
        def _blocking_input_loop():
            with self.term.cbreak(), self.term.raw():
                while self.should_run:
                    key = self.term.inkey(timeout=FPS)
                    if key in [chr(3)]:
                        self.should_run = False
                        break

                    self.on_tick(key)

        await asyncio.create_task(asyncio.to_thread(_blocking_input_loop))

    def on_tick(self, key):
        pass

    async def run(self):
        for renderer in self.render_list:
            renderer.initialize()

        await asyncio.gather(self.input_loop(), self.draw_loop())



import asyncio

from typing import Union

from a9s.aws_resources.route53 import Route53Table
from a9s.aws_resources.s3 import S3Table
from a9s.components.renderer import Renderer
from a9s.components.table import Table


class ServicesSelector(Renderer):
    def __init__(self, hud):
        super(ServicesSelector, self).__init__()
        self.services = {
            'route53': Route53Table,
            's3': S3Table,
        }

        self.hud = hud
        self._current_service = None

    def initialize(self):
        self.current_service = 'route53'

    def set_pos(self, *, x, y, to_x, to_y=None):
        if self._current_service:
            self._current_service.set_pos(x=x, y=y, to_x=to_x, to_y=to_y)

        super(ServicesSelector, self).set_pos(x=x, y=y, to_x=to_x, to_y=to_y)

    @property
    def current_service(self) -> Union[None, Table]:
        return self._current_service

    @current_service.setter
    def current_service(self, service):
        if service not in self.services:
            raise ValueError('Invalid service requested - {}!'.format(service))

        self._current_service = self.services[service]()
        self._current_service.initialize()
        self._current_service._echo_func = self._echo_func
        self._current_service.set_pos(x=self.x, y=self.y, to_x=self.to_x, to_y=self.to_y)
        self.hud.service = self._current_service

    def set_service(self, service):
        self.current_service = service

    def handle_key(self, key):
        if self.current_service:
            self.current_service.handle_key(key)

    async def update_data(self):
        to_await = [super(ServicesSelector, self).update_data()]
        if self.current_service:
            to_await.append(self.current_service.update_data())

        await asyncio.gather(*to_await)

    def onresize(self):
        if self.current_service:
            self.current_service.onresize()

    def fill_empty(self):
        if not self.current_service:
            return super(ServicesSelector, self).fill_empty()

        return self.current_service.fill_empty()

    def draw(self):
        if not self.current_service:
            return super(ServicesSelector, self).draw()

        self.current_service._echo_func = self._echo_func
        return self.current_service.draw()



from dataclasses import dataclass
from typing import Union, List
from copy import copy
from itertools import chain

from colored import attr


@dataclass
class Char:
    c: str
    fg: str = None
    bg: str = None
    should_reset_style: bool = False

    def __copy__(self):
        return Char(self.c, self.fg, self.bg, should_reset_style=self.should_reset_style)

    def __len__(self):
        return len(self.c)

    def format(self):
        to_ret = (self.fg or "") + (self.bg or "") + self.c
        if self.should_reset_style:
            to_ret += attr("reset")

        return to_ret

    def __add__(self, other):
        if isinstance(other, Char):
            return String([self, other])

        if isinstance(other, str):
            return String([self, Char(c=other)])

    def __repr__(self):
        return self.c

    def __str__(self):
        return self.format()


@dataclass
class Style:
    fg: str = None
    bg: str = None


class String:
    def __init__(self, starting_string: Union[str, "String", Char, List[Union["String", Char]]],
                 fg=None, bg=None):
        self._chars = []
        self.chars = starting_string
        self.fg = fg
        self.bg = bg

    def __len__(self):
        return sum(len(c) for c in self._chars)

    def join(self, str_list):
        if len(str_list) == 0:
            return String("")

        if len(str_list) == 1:
            return copy(str_list[0])

        total = copy(str_list[0])
        for elem in str_list[1:]:
            total += copy(self)
            total += copy(elem)

        return total

    def pad(self, width, char=" "):
        to_pad = width - len(self)
        if to_pad > 0:
            self._chars.append(String(char * to_pad))

        return self

    def reset_style_on_end(self):
        if len(self._chars) == 0:
            return self

        last = self._chars[-1]
        if isinstance(last, String):
            return last.reset_style_on_end()

        elif isinstance(last, Char):
            last.should_reset_style = True

        return self

    def with_style(self, style: Style):
        self.fg = style.fg
        self.bg = style.bg

        return self

    def find(self, seed):
        curr_i = 0
        was_matched = False
        initial_find = -1
        for i, c in enumerate(self):
            if repr(c) == seed[curr_i]:
                if not was_matched:
                    initial_find = i
                    was_matched = True
                curr_i += 1

            else:
                initial_find = -1
                curr_i = 0
                was_matched = False

            if curr_i == len(seed):
                return initial_find

        return -1

    def split(self, seed, count):
        if len(seed) == 0:
            return [self]

        j = 0
        i = 0
        times_splitted = 0

        splitted_str = list(self)
        length = len(splitted_str)
        if len(seed) > length:
            return [self]

        current_fragment = []
        fragments = [current_fragment]
        was_matched = False
        fragment_appended = True
        while i + j < length:
            if repr(splitted_str[i + j]) == seed[j] and times_splitted < count:
                if not was_matched:
                    if not fragment_appended:
                        fragments.append(current_fragment)
                    was_matched = True
                    current_fragment = []
                    fragment_appended = False

                current_fragment.append(splitted_str[i + j])
                j += 1
                if j == len(seed):  # we got a full match dismiss current fragment
                    current_fragment = []
                    times_splitted += 1
                    fragment_appended = False
                    was_matched = False
                    i += j
                    j = 0

            else:
                if was_matched:
                    fragments[-1].extend(current_fragment)
                    current_fragment = fragments[-1]  # there was no match, reuse old fragment
                    fragment_appended = True
                    was_matched = False
                    i += j
                    j = 0

                current_fragment.append(splitted_str[i + j])
                i += 1

        if not fragment_appended:
            fragments.append(current_fragment)

        return [String(f) for f in fragments]

    def force_self_style(self, fg=True, bg=False):
        to_ret = []
        for c in self:
            if fg:
                c.fg = self.fg

            if bg:
                c.bg = self.bg

            to_ret.append(c)

        return String(to_ret)

    @property
    def chars(self):
        return self._chars

    @chars.setter
    def chars(self, string):
        if isinstance(string, str):
            self._chars = [Char(c=c) for c in string]

        elif isinstance(string, list):
            self._chars = string

        else:
            raise ValueError("cannot set non Char / str instance")

    def __contains__(self, item):
        return item in repr(self)

    def __getitem__(self, item):
        chars = list(self)
        if isinstance(item, slice):
            to_ret = chars[item]
            return String(to_ret)

        char = chars[item]
        return char

    def __copy__(self):
        return String(copy(self._chars), self.fg, self.bg)

    def normal_iter(self):
        return iter(self._chars)

    def __iter__(self):
        all_iters = []

        for item in self._chars:
            item = copy(item)
            if item.bg is None:
                item.bg = self.bg

            if item.fg is None:
                item.fg = self.fg

            if isinstance(item, String):
                all_iters.append(iter(item))
                continue

            # item is Char class
            all_iters.append([item])

        return chain(*all_iters)

    def __add__(self, other):
        if isinstance(other, String):
            self._chars.append(other)

        if isinstance(other, Char):
            self._chars.append(other)

        if isinstance(other, str):
            self._chars.extend([Char(c=c) for c in other])

        return self

    def format(self):
        return (self.fg or "") + (self.bg or "") + "".join(c.format() for c in self._chars)

    def __repr__(self):
        return "".join(repr(c) for c in self._chars)

    def __str__(self):
        return self.format()



import curses
from dataclasses import dataclass

from a9s.components.custom_string import String


TEXT_OVERRIDE = {
    curses.KEY_EXIT: 'ESC',
    curses.KEY_ENTER: 'ENTER',
    curses.KEY_UP: 'UP',
    curses.KEY_DOWN: 'DOWN',
    curses.KEY_RIGHT: 'RIGHT',
    curses.KEY_LEFT: 'LEFT',
    curses.KEY_BACKSPACE: '⌫',
}


def format_special(key):
    if isinstance(key, int):
        return TEXT_OVERRIDE.get(key, key)

    return key


@dataclass
class KeySet:
    keys: tuple
    description: str = ''

    def __add__(self, other: 'KeySet'):
        return KeySet(other.keys + self.keys)

    def format(self, key_style):
        to_ret = String("")
        to_ret += String(", ".join([f'"{format_special(key)}"' for key in self.keys]), fg=key_style).reset_style_on_end()
        if self.description:
            to_ret += String(f" - {self.description}")

        return to_ret


BACK_KEYS = KeySet((curses.KEY_EXIT,), description='Cancel/Go back')
SEARCH_KEYS = KeySet(("/",), description='Search mode')
COMMAND_KEYS = KeySet((":",), description='Command mode')
ENTER_KEYS = KeySet((curses.KEY_ENTER,), description='Select')
DELETE_KEYS = KeySet((curses.KEY_BACKSPACE,))
PASTE_KEYS = KeySet(("\x16",),)  # ctrl + v
COPY_KEYS = KeySet(("y",), description='Yank mode')

LEFT_KEYS = KeySet(('h', curses.KEY_LEFT), description='Go left')
DOWN_KEYS = KeySet(('j', curses.KEY_DOWN), description='Go down')
UP_KEYS = KeySet(('k', curses.KEY_UP), description='Go up')
RIGHT_KEYS = KeySet(('l', curses.KEY_RIGHT), description='Go right')
END_OF_LINE_KEYS = KeySet(('$',), description='Go to end of line')
START_OF_LINE_KEYS = KeySet(('0',), description='Go to start of line')
END_OF_DOC_KEYS = KeySet(('G',), description='Go to bottom')
START_OF_DOC_KEYS = KeySet(('g',), description='Go to top')


def is_match(key, *key_sets: list[KeySet], code_only=False):
    return any(((key.code in key_set.keys) or (not code_only and (key in key_set.keys))) for key_set in key_sets)



from colored import fg

from a9s.components.custom_string import String
from a9s.components.renderer import Renderer

FRAMES = "⣾⣽⣻⢿⡿⣟⣯⣷"


class Loader(Renderer):
    MAX_FRAMES = len(FRAMES)

    def __init__(self):
        super(Loader, self).__init__()
        self.frame = 0

    def draw(self):
        if self.frame >= self.MAX_FRAMES:
            self.frame = 0

        current_frame = self.frame % len(FRAMES)
        self.echo(String(FRAMES[current_frame], fg=fg('light_green_3')))

        self.frame += 1




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['a9s', 'a9s.aws_resources', 'a9s.components']

package_data = \
{'': ['*']}

install_requires = \
['attrdict>=2.0.1,<3.0.0',
 'blessed>=1.18.0,<2.0.0',
 'boto3>=1.17.87,<2.0.0',
 'cached-property>=1.5.2,<2.0.0',
 'colored>=1.4.2,<2.0.0',
 'pyperclip>=1.8.2,<2.0.0']

entry_points = \
{'console_scripts': ['a9s = a9s.main:main']}

setup_kwargs = {
    'name': 'a9s',
    'version': '0.6.1',
    'description': 'Cli tool for navigation in Amazon AWS services. Highly inspired from k9s',
    'long_description': '# a9s\n\n![](https://img.shields.io/github/v/release/IamShobe/a9s) ![](https://img.shields.io/github/workflow/status/IamShobe/a9s/Create%20and%20publish%20a%20Python%20package?label=pypi%20build) ![](https://img.shields.io/github/workflow/status/IamShobe/a9s/Create%20and%20publish%20a%20Docker%20image?label=docker%20build)  \nCli tool for easily navigating in AWS services.  \nHighly inspired from [k9s](https://github.com/derailed/k9s). \n\n\n## How to install\n\n```shell\npip install a9s\n```\n\n### Docker build\n\n```shell\ndocker build . -t a9s\ndocker run -v ~/.aws/:/root/.aws -it --rm a9s\n```\n\n### Running docker from cloud\n\n```shell\ndocker run -v ~/.aws/:/root/.aws -it --rm ghcr.io/iamshobe/a9s\n```\n\n\n### How to develop\n\n#### Running mock server\nInstall poetry env:\n```bash\npoetry install\n```\nStart dev server:\n```bash\npoetry run moto_server -p 54321\n```\nRun mock data:\n```bash\npoetry run python -m mocked_env.main\n```\n\n#### Running mock server with docker-compose\n```bash\ndocker-compose -f mocked_env/docker-compose.yaml up --build\n```\n\n#### Running a9s in with mocked server\nRun a9s in local mode (connects to mock server on port 54321):\n```bash\nLOCAL=true poetry run a9s\n```\n\n## Goals\n\n### Services\n- [X] s3 support\n- [X] route53 support\n- [ ] EC2 support\n- [ ] ELB support\n- [ ] Cloudfront support\n\n\n### Features\n- [X] responsive tables\n- [X] allow to easily switch between services\n- [X] auto-complete commands\n- [X] vim shortcuts support\n- [X] opening files in S3\n- [X] quick yank\n- [ ] smart navigation between services - route53 pointing to ELB etc..\n',
    'author': 'Elran Shefer',
    'author_email': 'elran777@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/IamShobe/a9s',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import math
from enum import Enum
from colored import fg, bg

from a9s.components.custom_string import String, Style
from a9s.components.renderer import Renderer


class KeyMode(Enum):
    Navigation = 0
    Search = 1
    Command = 2
    Debug = 3


mode_to_str = {
    KeyMode.Navigation: "NAVIGATE",
    KeyMode.Search: "SEARCH",
    KeyMode.Command: "COMMAND",
    KeyMode.Debug: "D-NAV",
}

mode_to_color = {
    KeyMode.Navigation: bg("light_blue"),
    KeyMode.Search: bg("green"),
    KeyMode.Command: bg("light_red"),
    KeyMode.Debug: bg("yellow"),
}


class Mode(Renderer):
    def __init__(self):
        super(Mode, self).__init__()
        self.mode = KeyMode.Navigation

    def draw(self):
        style = Style(fg=fg("white"), bg=mode_to_color[self.mode])
        to_print = String(mode_to_str[self.mode]).with_style(style)
        spaces = (self.width - len(mode_to_str[self.mode]))
        left_side_spaces = String(math.floor(spaces / 2) * " ").with_style(style)
        right_side_spaces = String(math.ceil(spaces / 2) * " ").with_style(style)
        self.echo(left_side_spaces + to_print + right_side_spaces)



import math

from colored.colored import bg, fg, attr

from attrdict import AttrDict

from a9s.components.renderer import Renderer
from a9s.components.custom_string import String, Style


services = {
    'S3': {
        'colors': {
            'bg': 'red',
        }
    },
    'Route 53': {
        'colors': {
            'bg': 'orange_1'
        }
    }
}

class HUDComponent:
    SERVICE_NAME = None

    def get_hud_text(self, space_left):
        return String("")


class HUD(Renderer):
    SERVICE_SPACE = 10
    def __init__(self):
        super().__init__()
        self.service = None
    
    @property
    def service_props(self):
        return AttrDict(services[self.service.SERVICE_NAME])

    def draw(self):
        style = Style(fg=fg(self.service_props.colors.get('fg', 'white')) + attr('bold'), bg=bg(self.service_props.colors.bg))
        to_print = String(self.service.SERVICE_NAME).with_style(style)
        spaces = (self.SERVICE_SPACE - len(self.service.SERVICE_NAME))
        left_side_spaces = String(math.floor(spaces / 2) * " ").with_style(style)
        right_side_spaces = String(math.ceil(spaces / 2) * " ").with_style(style)
        self.echo(left_side_spaces + to_print + right_side_spaces, no_new_line=True)
        service_text = self.service.get_hud_text(self.width - self.SERVICE_SPACE)
        self.echo(service_text, x=self.x + self.SERVICE_SPACE, no_new_line=True)

        text_len = len(service_text)

        self.echo(String(" " * (self.width - self.SERVICE_SPACE - text_len)),
                  x=self.x + self.SERVICE_SPACE + text_len)



import asyncio
from dataclasses import dataclass
from typing import Callable

import signal
import pyperclip

from a9s.aws_resources.loader import Loader
from a9s.components.keys import is_match, SEARCH_KEYS, COMMAND_KEYS, ENTER_KEYS, DELETE_KEYS, PASTE_KEYS, BACK_KEYS
from a9s.components.renderer import ScrollableRenderer
from a9s.aws_resources.help import Help
from a9s.aws_resources.logo import Logo
from a9s.aws_resources.hud import HUD
from a9s.aws_resources.services import ServicesSelector
from a9s.components.app import App
from a9s.components.logger import logger
from a9s.components.mode import KeyMode, Mode
from a9s.components.autocomplete import AutoComplete
from a9s import __version__


@dataclass
class Context:
    mode: KeyMode
    focused: Callable[[], ScrollableRenderer]

    def __eq__(self, other):
        return self.mode == other.mode and self.focused() == other.focused()


class MainApp(App):
    def __init__(self):
        self.logo = Logo(__version__)
        self.help = Help()
        self.hud = HUD()
        self.services_selector = ServicesSelector(hud=self.hud)
        self.mode_renderer = Mode()
        self.loader = Loader()
        self.auto_complete = AutoComplete(commands=self.commands)

        super(MainApp, self).__init__([
            self.logo, self.services_selector, self.mode_renderer, self.auto_complete, logger, self.hud, self.help, self.loader,
        ])
        self.context_stack = [Context(mode=KeyMode.Navigation, focused=lambda: self.services_selector.current_service)]

    @property
    def commands(self):
        return {
            'debug': self.on_debug_command,
            'hideDebug': self.on_hide_debug_command,
            'quit': self.on_quit_command,

            's3': lambda: self.services_selector.set_service('s3'),
            'route53': lambda: self.services_selector.set_service('route53'),
        }

    def on_debug_command(self):
        current_context = self.context_stack[-1]
        debug_context = Context(mode=KeyMode.Navigation, focused=lambda: logger)
        if current_context != debug_context:
            self.context_stack.append(debug_context)

        logger.halt_debug()
        logger.show()

    def on_hide_debug_command(self):
        current_context = self.context_stack[-1]
        if current_context.focused() == logger:
            self.context_stack.pop()

        logger.hide()

    def on_quit_command(self):
        self.should_run = False

    def on_resize(self, *args):
        self.help.set_pos(x=25, to_x=self.term.width, y=0, to_y=8)
        self.logo.set_pos(x=3, y=0, to_x=24, to_y=6)
        self.hud.set_pos(x=0, y=9, to_x=self.term.width - 1)
        self.loader.set_pos(x=self.term.width-1, y=9, to_x=self.term.width)
        self.mode_renderer.set_pos(x=0, y=self.term.height - 1, to_x=10)
        self.auto_complete.set_pos(x=11, y=self.term.height - 1, to_x=self.term.width)
        self.services_selector.set_pos(x=0, y=10, to_x=self.term.width, to_y=self.term.height - 1)
        logger.set_pos(x=max(self.term.width - 40, 0), y=0, to_x=self.term.width, to_y=10)

        super(MainApp, self).on_resize()

    def on_tick(self, key):
        current_context = self.context_stack[-1]
        self.mode_renderer.mode = current_context.mode
        focused = current_context.focused()
        self.loader.shown = focused.data_updating
        self.help.yank_mode = focused.yank_mode
        self.auto_complete.guess_mode = current_context.mode
        if key:
            logger.debug("pressed key: " + repr(key))

        if current_context.mode == KeyMode.Navigation:
            should_stop_propagate = focused.handle_key(key)
            self.auto_complete.text = ("/" + focused.filter) if focused.filter else ""
            if is_match(key, BACK_KEYS) and not should_stop_propagate:
                if len(self.context_stack) > 1:
                    self.context_stack.pop()
                    logger.debug("Popping context")

            if is_match(key, SEARCH_KEYS):
                self.context_stack.append(Context(mode=KeyMode.Search, focused=current_context.focused))
                logger.debug("Switching to Search mode")
            elif is_match(key, COMMAND_KEYS):
                self.context_stack.append(Context(mode=KeyMode.Command, focused=current_context.focused))
                logger.debug("Switching to Command mode")

            if is_match(key, SEARCH_KEYS, COMMAND_KEYS):
                self.auto_complete.text = str(key)

        if current_context.mode in [KeyMode.Search, KeyMode.Command]:
            if is_match(key, BACK_KEYS):
                self.context_stack.pop()
                logger.debug("Popping context and switching to Navigate mode")

            if is_match(key, PASTE_KEYS):
                self.auto_complete.text += pyperclip.paste().replace("\n", "").replace("\r", "")

            elif is_match(key, DELETE_KEYS):
                self.auto_complete.delete_char()

            elif is_match(key, ENTER_KEYS):
                self.context_stack.pop()
                logger.debug("Popping context and switching to Navigate mode")
                if self.auto_complete.get_actual_text() == "":
                    self.auto_complete.text = ""

                if current_context.mode == KeyMode.Command:
                    self.auto_complete.execute_command_if_matched()

            elif key.isprintable():
                self.auto_complete.text += key

            if current_context.mode == KeyMode.Search:
                focused.filter = self.auto_complete.get_actual_text()

            if current_context.mode == KeyMode.Command:
                self.auto_complete.handle_key(key)

        if current_context.mode == KeyMode.Command:
            if is_match(key, ENTER_KEYS):
                self.auto_complete.text = ""


def main():
    app = MainApp()
    try:
        signal.signal(signal.SIGWINCH, app.on_resize)

    except Exception:
        pass

    asyncio.run(app.run())


if __name__ == '__main__':
    main()
